#pragma once
/************************************************************************
 * This file is part of the AREG SDK core engine.
 * AREG SDK is dual-licensed under Free open source (Apache version 2.0
 * License) and Commercial (with various pricing models) licenses, depending
 * on the nature of the project (commercial, research, academic or free).
 * You should have received a copy of the AREG SDK license description in LICENSE.txt.
 * If not, please contact to info[at]aregtech.com
 *
 * \copyright   (c) 2017-2021 Aregtech UG. All rights reserved.
 * \file        areg/component/private/ServiceManagerEvents.hpp
 * \ingroup     AREG SDK, Asynchronous Event Generator Software Development Kit 
 * \author      Artak Avetyan
 * \brief       AREG Platform, Service Manager communication Events
 *
 ************************************************************************/

/************************************************************************
 * Includes
 ************************************************************************/
#include "areg/base/GEGlobal.h"
#include "areg/component/TEEvent.hpp"
#include "areg/base/SharedBuffer.hpp"

/************************************************************************
 * Dependencies
 ************************************************************************/
class ProxyAddress;
class StubAddress;
class Channel;
class String;

//////////////////////////////////////////////////////////////////////////
// ServiceManagerEventData class declaration
//////////////////////////////////////////////////////////////////////////

/**
 * \brief   Service Manager event data object, which contains routing connection
 *          commands to execute and appropriate data for command
 **/
class ServiceManagerEventData
{
//////////////////////////////////////////////////////////////////////////
// Internal types and constants.
//////////////////////////////////////////////////////////////////////////
public:
    /**
     * \brief   The Service Manager event commands
     **/
    typedef enum class E_ServiceManagerCommands
    {
          CMD_StopRoutingClient         //!< Requested to stop Service Manager client connection
        , CMD_RegisterProxy             //!< Requested to register Proxy
        , CMD_UnregisterProxy           //!< Requested to unregister Proxy
        , CMD_RegisterStub              //!< Requested to register Stub
        , CMD_UnregisterStub            //!< Requested to unregister Stub
        , CMD_ConfigureConnection       //!< Requested to configure connection
        , CMD_StartConnection           //!< Requested to start connection, the data is configuration file
        , CMD_StartNetConnection        //!< Requested to start network connection, the data is IP-address and port number
        , CMD_StopConnection            //!< Requested to stop connection
        , CMD_SetEnableService          //!< Requested to enables or disables remote service
        , CMD_RegisterConnection        //!< Requested to register connection
        , CMD_UnregisterConnection      //!< Requested to unregister connection
        , CMD_LostConnection            //!< Requested to send notifications to components that connection is lost.
        , CMD_ShutdownService           //!< Request to shutdown service.
    } eServiceManagerCommands;

    /**
     * \brief   Converts event command to string
     * \param   val     The command to convert to string
     **/
    static const char * getString( ServiceManagerEventData::eServiceManagerCommands val );

//////////////////////////////////////////////////////////////////////////
// Static members
//////////////////////////////////////////////////////////////////////////
public:
    /**
     * \brief   Creates and returns Service Manager event data with command to stop router client connection.
     **/
    static ServiceManagerEventData stopMessageRouterClient( void );

    /**
     * \brief   creates and returns service manager event with command to stop routing connection and stop service manager.
     **/
    static ServiceManagerEventData shutdownServiceManager( void );

    /**
     * \brief   Creates and returns Service Manager event data with command to register Proxy
     **/
    static ServiceManagerEventData registerProxy( const ProxyAddress & addrProxy );

    /**
     * \brief   Creates and returns Service Manager event data with command to unregister Proxy
     **/
    static ServiceManagerEventData unregisterProxy( const ProxyAddress & addrProxy );

    /**
     * \brief   Creates and returns Service Manager event data with command to register Stub
     **/
    static ServiceManagerEventData registerStub( const StubAddress & addrStub );

    /**
     * \brief   Creates and returns Service Manager event data with command to unregister Stub
     **/
    static ServiceManagerEventData unregisterStub( const StubAddress & addrStub );

    /**
     * \brief   Creates and returns Service Manager event data with command to configure connection
     **/
    static ServiceManagerEventData configureConnection( const String & configFile );

    /**
     * \brief   Creates and returns Service Manager event data with command to start router client connection
     **/
    static ServiceManagerEventData startConnection( const String & configFile );

    /**
     * \brief   Creates and returns Service Manager event data with command to start router client connection.
     * \param   ipAddress   The IP-Address of remote Routing Service.
     * \param   portNr      The port number to connect.
     **/
    static ServiceManagerEventData startNetConnection( const String & ipAddress, unsigned short portNr );

    /**
     * \brief   Creates and returns Service Manager event data with command to stop connection
     **/
    static ServiceManagerEventData stopConnection( void );

    /**
     * \brief   Creates Service Manager event data with command to enable or disable remote service.
     *          If disabled, services cannot be accessed outside of process.
     *          If remote service is already running and it is requested to disable, it will be stopped.
     * \param  enable   If true, the remote services are accessed from outside of process.
     *                  If false, all services run and accessed only within process.
     * \return  Returns Service Manager Event data object with command to enable or disable remote service.
     **/
    static ServiceManagerEventData enableRemoteService( bool enable );

    /**
     * \brief   Creates and returns Service Manager event data with command to register connection
     * \param   channel     The Channel information to set for registered connection
     **/
    static ServiceManagerEventData registerConnection( const Channel & channel );

    /**
     * \brief   Creates and returns Service Manager event data with command to unregister connection
     * \param   channel     The Channel information to set for unregistered connection
     **/
    static ServiceManagerEventData unregisterConnection( const Channel & channel );

    /**
     * \brief   Creates and returns Service Manager event data with command of lost connection
     * \param   channel     The Channel information to set for lost connection
     **/
    static ServiceManagerEventData lostConnection( const Channel & channel );

//////////////////////////////////////////////////////////////////////////
// Constructors / Destructor
//////////////////////////////////////////////////////////////////////////
public:
    /**
     * \brief   Constructor. Sets Service Manager event command
     * \param   cmdServiceManager    The command to set
     **/
    inline explicit ServiceManagerEventData( ServiceManagerEventData::eServiceManagerCommands cmdServiceManager );

    /**
     * \brief   Copy constructor.
     * \param   src     The source of data to copy.
     **/
    inline ServiceManagerEventData( const ServiceManagerEventData & source );

    /**
     * \brief   Move constructor.
     * \param   src     The source of data to move.
     **/
    inline ServiceManagerEventData( ServiceManagerEventData && source ) noexcept;

    /**
     * \brief   Destructor.
     **/
    ~ServiceManagerEventData( void ) = default;

//////////////////////////////////////////////////////////////////////////
// Attributes
//////////////////////////////////////////////////////////////////////////
public:

    /**
     * \brief   Copies Service Manager event data with comment from given source.
     * \param   source  The source of data to copy
     **/
    inline ServiceManagerEventData & operator = ( const ServiceManagerEventData & source );

    /**
     * \brief   Moves Service Manager event data with comment from given source.
     * \param   source  The source of data to move.
     **/
    inline ServiceManagerEventData & operator = ( ServiceManagerEventData && source ) noexcept;

    /**
     * \brief   Returns streaming object to read data
     **/
    inline const IEInStream & getReadStream( void ) const;

    /**
     * \brief   Returns streaming object to write data
     **/
    inline IEOutStream & getWriteStream( void );

    /**
     * \brief   Returns Service Manager command
     **/
    inline ServiceManagerEventData::eServiceManagerCommands getCommand( void ) const;

//////////////////////////////////////////////////////////////////////////
// Member variables
//////////////////////////////////////////////////////////////////////////
private:
    eServiceManagerCommands mCommand;   //!< Service Manager event data command
    SharedBuffer            mStream;    //!< Service Manager event data streaming object to write and read data.

//////////////////////////////////////////////////////////////////////////
// Forbidden calls
//////////////////////////////////////////////////////////////////////////
private:
    ServiceManagerEventData( void ) = delete;
};

//////////////////////////////////////////////////////////////////////////
// Service Manager event and consumer declaration
//////////////////////////////////////////////////////////////////////////
DECLARE_EVENT(ServiceManagerEventData, ServiceManagerEvent, IEServiceManagerEventConsumer)

//////////////////////////////////////////////////////////////////////////
// ServiceManagerEventData inline methods
//////////////////////////////////////////////////////////////////////////

inline ServiceManagerEventData::ServiceManagerEventData( ServiceManagerEventData::eServiceManagerCommands cmdServiceManager )
    : mCommand  ( cmdServiceManager )
    , mStream   ( )
{
}

inline ServiceManagerEventData::ServiceManagerEventData( const ServiceManagerEventData & source )
    : mCommand  ( source.mCommand )
    , mStream   ( source.mStream )
{
}

inline ServiceManagerEventData::ServiceManagerEventData( ServiceManagerEventData && source ) noexcept
    : mCommand  ( std::move(source.mCommand) )
    , mStream   ( std::move(source.mStream) )
{
}

inline ServiceManagerEventData & ServiceManagerEventData::operator = (const ServiceManagerEventData & source)
{
    mCommand= source.mCommand;
    mStream = source.mStream;

    return (*this);
}

inline ServiceManagerEventData & ServiceManagerEventData::operator = ( ServiceManagerEventData && source ) noexcept
{
    mCommand= source.mCommand;
    mStream = std::move(source.mStream);

    return (*this);
}

inline const IEInStream & ServiceManagerEventData::getReadStream( void ) const
{
    return static_cast<const IEInStream &>(mStream);
}

inline IEOutStream & ServiceManagerEventData::getWriteStream( void )
{
    return static_cast<IEOutStream &>(mStream);
}

inline ServiceManagerEventData::eServiceManagerCommands ServiceManagerEventData::getCommand( void ) const
{
    return mCommand;
}

inline const char * ServiceManagerEventData::getString( ServiceManagerEventData::eServiceManagerCommands val )
{
    switch ( val )
    {
    case ServiceManagerEventData::eServiceManagerCommands::CMD_StopRoutingClient:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_StopRoutingClient";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_RegisterProxy:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_RegisterProxy";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_UnregisterProxy:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_UnregisterProxy";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_RegisterStub:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_RegisterStub";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_UnregisterStub:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_UnregisterStub";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_ConfigureConnection:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_ConfigureConnection";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_StartConnection:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_StartConnection";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_StartNetConnection:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_StartNetConnection";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_StopConnection:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_StopConnection";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_SetEnableService:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_SetEnableService";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_RegisterConnection:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_RegisterConnection";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_UnregisterConnection:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_UnregisterConnection";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_LostConnection:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_LostConnection";
    case ServiceManagerEventData::eServiceManagerCommands::CMD_ShutdownService:
        return "ServiceManagerEventData::eServiceManagerCommands::CMD_ShutdownService";
    default:
        return "ERR: undefined ServiceManagerEventData::eServiceManagerCommands value!!!";
    }
}
